(* Solver *)

open Debug;;
open Problem;;

module SM = Map.Make(String);;

(*** Numbering *)
module Numbering =
  struct
    type 'a t = {
      mutable number : int;
      forward : ('a,int) Hashtbl.t;
      backward : (int,'a) Hashtbl.t
    };;
    let create ?(m=1024) () =
      { number = 0;
        forward = Hashtbl.create m;
        backward = Hashtbl.create m }
    ;;
    let number numbering = numbering.number;;
    let name numbering id = Hashtbl.find numbering.backward id;;
    let register numbering entry =
      try
        Hashtbl.find numbering.forward entry
      with
      | Not_found ->
          let id = numbering.number in
          numbering.number <- id + 1;
          Hashtbl.add numbering.forward entry id;
          Hashtbl.add numbering.backward id entry;
          id
    ;;
  end
;;
(* ***)

module IM = Map.Make(struct type t = int let compare = compare end);;
module IS = Set.Make(struct type t = int let compare = compare end);;

type clause = IS.t * IS.t

module CS = Set.Make(struct type t = clause let compare = compare end);;

exception Lose;;
exception Win;;

type env = {
  clauses : CS.t;
  assignments : (int * bool) list;
  depth : int
}

type polarity =
| Positive
| Negative
| Bipolar

let fp = Printf.fprintf;;

let dump_human_readable oc variables clauses =
  CS.iter
    begin fun (pos,neg) ->
      IS.iter (fun variable -> fp oc " !%s" (Numbering.name variables variable)) neg;
      IS.iter (fun variable -> fp oc " %s" (Numbering.name variables variable)) pos;
      fp oc "\n"
    end
    clauses
;;

let dump_machine_readable oc variables clauses =
  fp oc "p cnf %d %d\n" (Numbering.number variables) (CS.cardinal clauses);
  CS.iter
    begin fun (pos,neg) ->
      IS.iter (fun variable -> fp oc " %d" (- (variable + 1))) neg;
      IS.iter (fun variable -> fp oc " %d" (variable + 1)) pos;
      fp oc " 0\n"
    end
  clauses
;;

let solve problem =
  (* Number variables *)
  let variables = Numbering.create () in
  let clauses = ref CS.empty in
  let string_of_clause (pos,neg) =
    sf "(%s,%s)"
      (String.concat " " (List.map (Numbering.name variables) (IS.elements pos)))
      (String.concat " " (List.map (Numbering.name variables) (IS.elements neg)))
  in
  (*** Construct hashtables *)
  List.iter
    begin fun (clause, reasons) ->
      let clause' =
        List.fold_left
          begin fun (pos,neg) (polarity, variable) ->
            let v = Numbering.register variables variable in
            if polarity then
              (IS.add v pos, neg)
            else
              (pos, IS.add v neg)
          end
          (IS.empty, IS.empty)
          clause
      in
      clauses := CS.add clause' !clauses
    end
    problem;
  (* ***)
  (*debug (sf "Total %d clauses and %d variables" (CS.cardinal !clauses) (Numbering.number variables));*)
  begin
    match !Opt.dump_clauses with
    | None -> ()
    | Some fn ->
      let oc = open_out fn in
      if !Opt.cnf then
        dump_machine_readable oc variables !clauses
      else
        dump_human_readable oc variables !clauses;
      close_out oc
  end;
  (*let variable_to_clause_map env =
    let register_clause variable clause vcm =
      IM.add
        variable
          (CS.add clause
            (try
               IM.find variable vcm
            with
            | Not_found -> CS.empty))
    in
    CS.fold
      begin fun ((pos, neg) as clause) vcm ->
        let vcm =
          IS.fold
            begin fun variable vcm ->
              register_variable variable clause vcm
            end
            pos
            vcm
        in
        let vcm =
          IS.fold
            begin fun variable vcm ->
              register_variable variable clause vcm
            end
            neg
            vcm
        in
        vcm
      end
      env.clauses
  in*)
  let assign variable sign env =
    debug2 10 df "Setting %s to %b" (Numbering.name variables variable) sign;
    { 
      depth = env.depth + 1;
      assignments = (variable,sign) :: env.assignments;
      clauses =
        CS.fold
          begin fun ((pos, neg) as clause) clauses ->
            if sign then
              if IS.mem variable pos then
                (* Clause becomes true *)
                clauses
              else
                CS.add (pos, IS.remove variable neg) clauses
            else
              if IS.mem variable neg then
                clauses
              else
                CS.add (IS.remove variable pos, neg) clauses
          end
          env.clauses
          CS.empty }
  in
  let choose_clause env =
    (* Select one of the smallest clause *)
    let (_, clause) =
      CS.fold
        begin fun ((pos, neg) as clause) (best_size, best) ->
          let size = (IS.cardinal pos) + (IS.cardinal neg) in
          if size < best_size then
            (size, clause)
          else
            (best_size, best)
        end
        env.clauses
        (max_int, (IS.empty, IS.empty))
    in
    clause
  in
  let rec solve env =
    (*debug (sf "Depth %d, %d clauses" env.depth (CS.cardinal env.clauses));*)
    let env = perform_forced env in
    let env = remove_satisfied env in
    if CS.is_empty env.clauses then
      begin
        (*debug (sf "No more clauses, formula satisfied");*)
        env
      end
    else
      (* Check if we have the empty clause *)
      if CS.mem (IS.empty, IS.empty) env.clauses then
        begin
          (*debug (sf "Empty clause, backtracking (depth %d)" env.depth);*)
          raise Lose
        end
      else
        begin
          (* Chose a clause to satisfy *)
          let (pos, neg) = choose_clause env in
          let (variable, polarity) =
            if IS.is_empty pos then
              (IS.choose neg, false)
            else
              (IS.choose pos, true)
          in
          try
            solve (find_sequents (check_unipolar (assign variable polarity env)))
          with
          | Lose ->
            (* Try to assign variable to true *)
            (*debug "Bad guess";*)
            solve (find_sequents (check_unipolar (assign variable (not polarity) env)))
        end
  and perform_forced env =
    let env = 
      CS.fold
        begin fun ((pos,neg) as clause) env ->
          if IS.cardinal pos = 1 && IS.is_empty neg then
            assign (IS.choose pos) true env
          else
            if IS.is_empty pos && IS.cardinal neg = 1 then
              assign (IS.choose neg) false env
            else
              env
        end
        env.clauses
        env
    in
    env
  and remove_satisfied env =
    (* Check if we have satisfied clauses *)
    let clauses =
      CS.fold
        begin fun ((pos,neg) as clause) clauses ->
          if IS.is_empty (IS.inter pos neg) then
            CS.add clause clauses
          else
            (* Clause has both a literal and its negation, remove it *)
            begin
              (*debug (sf "Clause %s removed because it has both a literal and its negation" (string_of_clause clause)); *)
              clauses
            end
        end
        env.clauses
        CS.empty
    in
    { env with clauses = clauses }
  and sequent_counter = ref 0
  and find_sequents env =
    env
    (*
    (*incr sequent_counter;
    if !sequent_counter <> 50 then
      env
    else*)
      begin
        sequent_counter := 0;
    let (positives, negatives) =
      CS.fold
        begin fun ((pos,neg) as clause) (positives, negatives) ->
          let positives =
            IS.fold
              begin fun variable positives ->
                IM.add
                  variable
                  (CS.add
                    clause
                    (try
                      IM.find variable positives
                    with
                    | Not_found -> CS.empty))
                  positives
              end
              pos
              positives
          and negatives =
            IS.fold
              begin fun variable negatives ->
                IM.add
                  variable
                  (CS.add
                    clause
                    (try
                      IM.find variable negatives
                    with
                    | Not_found -> CS.empty))
                  negatives
              end
              neg
              negatives
          in
          (positives, negatives)
        end
        env.clauses
        (IM.empty, IM.empty)
    in
    (* Find a variable for which two oppositely containing clauses exist *)
    (* We are not supposed to have unipolar variables ! *)
    IM.fold
      begin fun variable variant_clauses env ->
        let contravariant_clauses = try IM.find variable negatives with Not_found -> CS.empty in
        if CS.is_empty variant_clauses or CS.is_empty contravariant_clauses then
          env
        else
          let (pos1,neg1) = CS.choose variant_clauses
          and (pos2,neg2) = CS.choose contravariant_clauses
          in
          let ((pos',neg') as clause') =
            (IS.remove variable (IS.union pos1 pos2), IS.remove variable (IS.union neg1 neg2))
          in
          if IS.is_empty (IS.inter pos' neg') then
            (* Generate a new clause *)
            begin
              (*debug (sf "Adding new clause %s" (string_of_clause clause'));*)
              { env with clauses = CS.add clause' env.clauses }
            end
          else
            env
      end
      positives
      env
      end*)
  and check_unipolar env =
    (*debug (sf "Depth %d, checking for unipolar variables." env.depth);*)
    let (positives, negatives) =
      CS.fold
        begin fun (pos,neg) (positives, negatives) ->
          (IS.union pos positives, IS.union neg negatives)
        end
        env.clauses
        (IS.empty, IS.empty)
    in
    let unipolar_positives = IS.diff positives negatives
    and unipolar_negatives = IS.diff negatives positives
    in
    let env =
      IS.fold
        begin fun variable env ->
          (*debug (sf "Variable %s is unipolar positive." (Numbering.name variables variable));*)
          assign variable true env
        end
        unipolar_positives
        env
    in
    let env =
      IS.fold
        begin fun variable env ->
          (*debug (sf "Variable %s is unipolar negative." (Numbering.name variables variable));*)
          assign variable false env
        end
        unipolar_negatives
        env
    in
    env
  in
  try
    let solution = solve { assignments = []; depth = 0; clauses = !clauses } in
    let solution' =
      List.fold_left
        begin fun result (variable,assignment) ->
          if assignment then
            Numbering.name variables variable :: result
          else
            result
        end
        []
        solution.assignments
    in
    Some(Set({
      ssl_subset = solution';
      ssl_cost = 0.0
    }))
  with
  | Lose -> None
;;
